unit EloTextBox;

interface

uses
  SysUtils, Classes, Controls, Messages, Windows, StdCtrls, Dialogs, Graphics;

type
  TAlignmentText = (atNone, atWidth, atWidthAndHyphen);
  TOEMConvert = (tcNone, tcOemToAnsi, tcAnsiToOem);
  TOnDrawCurrentLine = procedure(aCanvas: TCanvas; X, Y: Integer;
    aRect: TRect; var Text: String; var DefaultDraw: Boolean) of object;
  TOnErrorTextFormat = procedure(Sender: TObject; aCanvas: TCanvas;
    var aWord: String) of object;

  TCustomEloTextBox = class(TCustomControl)
  private
    { Private declarations }
    FAlignmentText: TAlignmentText;
    FStrings: TStrings;
    FTabString: String;
    FFileName: TFileName;
    FVScrollBar: Boolean;
    FSpacing: Byte;
    FOEMConvert: TOEMConvert;
    FHighlightColor, FHighlightTextColor: TColor;
    FHotTrack: Boolean;
    FBOFText, FEOFText: Boolean;
    FOnVScroll: TScrollEvent;
    FOnDrawCurrentLine: TOnDrawCurrentLine;
    FOnPaint: TNotifyEvent;
    FOnErrorTextFormat: TOnErrorTextFormat;
    procedure SetAlignmentText(Value: TAlignmentText);
    procedure SetStrings(Value: TStrings);
    procedure SetTabString(Value: String);
    procedure SetFileName(Value: TFileName);
    procedure SetVScrollBar(Value: Boolean);
    procedure SetOEMConvert(Value: TOEMConvert);
    procedure SetSpacing(Value: Byte);
    procedure SetHighlightColor(Value: TColor);
    procedure SetHighlightTextColor(Value: TColor);
    procedure SetHotTrack(Value: Boolean);
    function GetCurHeight: Integer;
    function GetCurWidth: Integer;
    procedure SetTextScrollRange(HWND: THandle; nBar, nMinPos, nMaxPos: Integer;
      bRedraw: Boolean);
    procedure WMErasebkgnd(var Message: TWMErasebkgnd); message WM_ERASEBKGND;
    procedure WMWindowPosChanged(var Message: TWMWindowPosChanged); message WM_WINDOWPOSCHANGED;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure WMNCPaint(var Message: TMessage); message WM_NCPAINT;
    procedure WMSetFont(var Message: TWMSetFont); message WM_SETFONT;
    procedure WMVScroll(var Message: TWMVScroll); message WM_VSCROLL;
    procedure WMEnable(var Message: TWMEnable); message WM_ENABLE;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure CNKeyDown(var Message: TWMKeyDown); message CN_KEYDOWN;
  protected
    { Protected declarations }
    function RusHyps(Wd: String): String;
    function Hyphen(Wd: String): String;
    function GetWordPart(Wd: String; ML: Integer; EmptyAllowed: Boolean; var J: Integer): String;
    function CheckString(strSource: String): String;
    procedure MakeList(List: TStringList);
    procedure LoadFromFile(FileName: TFileName);
    procedure LoadFromList(List: TStrings);
    procedure DrawCurrentLine(aCanvas: TCanvas; X, Y: Integer;
      aRect: TRect; var Text: String);
    property BevelWidth;
  public
    { Public declarations }
    WorkList: TStrings;
    RangeBlockingMouseMove: Cardinal; //    MouseMove
    constructor Create(AOwner: TComponent); override;
    procedure CreateParams(var Params: TCreateParams); override;
    destructor Destroy; override;
    procedure DefaultHandler(var Message); override;
    procedure Paint; override;
    procedure Clear;
    procedure SetTextVScrollPos(HWND: THandle; nBar, nPos: Integer;
      bRedraw: Boolean);
    procedure ResetPosParams;
    procedure BlockingMouseMove;
    procedure TextHome;
    procedure TextPageUp;
    procedure TextPrior;
    procedure TextNext;
    procedure TextPageDown;
    procedure TextEnd;
    property VScrollBar: Boolean read FVScrollBar default False;
    { Published declarations }
    property AlignmentText: TAlignmentText read FAlignmentText write SetAlignmentText default atWidth;
    property Strings: TStrings read FStrings write SetStrings;
    property FileName: TFileName read FFileName write SetFileName;
    property TabString: String read FTabString write SetTabString;
    property OEMConvert: TOEMConvert read FOEMConvert write SetOEMConvert default tcNone;
    property Spacing: Byte read FSpacing write SetSpacing default 4;
    property HighlightColor: TColor read FHighlightColor write SetHighlightColor default clHighlight;
    property HighlightTextColor: TColor read FHighlightTextColor write SetHighlightTextColor default clHighlightText;
    property HotTrack: Boolean read FHotTrack write SetHotTrack default True;
    property BOFText: Boolean read FBOFText;
    property EOFText: Boolean read FEOFText;
    property OnVScroll: TScrollEvent read FOnVScroll write FOnVScroll;
    property OnDrawCurrentLine: TOnDrawCurrentLine read FOnDrawCurrentLine write FOnDrawCurrentLine;
    property OnPaint: TNotifyEvent read FOnPaint write FOnPaint;
    property OnErrorTextFormat: TOnErrorTextFormat read FOnErrorTextFormat write FOnErrorTextFormat;
  end;

  TEloStringsBox = class(TStringList)
  private
    { Private declarations }
    EloTextBox: TCustomEloTextBox;
    NoChanged: Boolean;
  protected
    { Protected declarations }
    procedure Changed; override;
  public
    { Public declarations }
    constructor Create;
    procedure LoadFromStream(Stream: TStream); override;
    procedure Assign(Source: TPersistent); override;
  end;

  TEloTextBox = class(TCustomEloTextBox)
  published
    { Published declarations }
    property Align;
    property Anchors;
    property Color default clWindow;
    property Constraints;
    property Hint;
    property ShowHint;
    property Visible;
    property Enabled;
    property Font;
    property ParentFont;
    property PopupMenu;

    property BevelEdges;
    property BevelInner default bvNone;
    property BevelOuter default bvLowered;
    property BevelKind default bkSoft;
    property BorderWidth;
    property Ctl3D;
    property DockSite;
    property DockManager;
    property DragCursor;
    property DragKind;
    property DragMode;
//   property ImeMode;
//   property ImeName;
    property ParentCtl3D;
    property TabOrder;
    property TabStop;

    property OnClick;
    property OnDblClick;
    property OnEnter;
    property OnExit;

    property AlignmentText;
    property Strings;
    property FileName;
    property TabString;
    property OEMConvert;
    property Spacing;
    property HighlightColor;
    property HighlightTextColor;
    property HotTrack;

    property OnGetSiteInfo;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnUnDock;
    property OnStartDock;
    property OnStartDrag;

    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;

    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;

    property OnVScroll;
    property OnDrawCurrentLine;
    property OnPaint;
    property OnErrorTextFormat;
  end;


procedure Register;

implementation

var
   ParagraphArray: array of String;
   BeginScrollPos: Integer = 0;
   VisibleLineCount: Integer = 0;
   LineHeight: Integer = 0;
   CurrentLine: Integer = -1;
   OldLine: Integer = -1;
   PaintAllLines: Boolean = True;
   NoMouseMove: Boolean = False;
   BeginBlockingMouseMove: Cardinal = 0;
   EventMouseLeave: Boolean = False;
   IndexLineClear: Integer = -1;

procedure Register;
begin
  RegisterComponents('EloSoft', [TEloTextBox]);
end;

function KeyDataToShiftState(KeyData: Longint): TShiftState;
const
  AltMask = $20000000;
{$IFDEF LINUX}
  CtrlMask = $10000000;
  ShiftMask = $08000000;
{$ENDIF}
begin
  Result := [];
  if GetKeyState(VK_SHIFT) < 0 then Include(Result, ssShift);
  if GetKeyState(VK_CONTROL) < 0 then Include(Result, ssCtrl);
  if KeyData and AltMask <> 0 then Include(Result, ssAlt);
{$IFDEF LINUX}
  if KeyData and CtrlMask <> 0 then Include(Result, ssCtrl);
  if KeyData and ShiftMask <> 0 then Include(Result, ssShift);
{$ENDIF}
end;

function ConvertOemToAnsi(const S: String): String;
{ ConvertOemToAnsi translates a string from the OEM-defined
  character set into either an ANSI or a wide-character string }
{$IFNDEF WIN32}
var
  Source, Dest : array[0..255] of Char;
{$ENDIF}
begin
{$IFDEF WIN32}
  SetLength(Result, Length(S));
  if Length(Result) > 0 then
    OemToAnsi(PChar(S), PChar(Result));
{$ELSE}
  if Length(Result) > 0 then
  begin
    OemToAnsi(StrPCopy(Source, S), Dest);
    Result := StrPas(Dest);
  end;
{$ENDIF}
end; { ConvertOemToAnsi }

function ConvertAnsiToOem(const S: String): String;
{ ConvertAnsiToOem translates a string into the OEM-defined character set }
{$IFNDEF WIN32}
var
  Source, Dest : array[0..255] of Char;
{$ENDIF}
begin
{$IFDEF WIN32}
  SetLength(Result, Length(S));
  if Length(Result) > 0 then
    AnsiToOem(PChar(S), PChar(Result));
{$ELSE}
  if Length(Result) > 0 then
  begin
    AnsiToOem(StrPCopy(Source, S), Dest);
    Result := StrPas(Dest);
  end;
{$ENDIF}
end; { ConvertAnsiToOem }

{ TEloStringsBox }

constructor TEloStringsBox.Create;
begin
  inherited Create;
  NoChanged:= False;  
end;

procedure TEloStringsBox.Changed;
begin
  inherited Changed;
  if not NoChanged and EloTextBox.HandleAllocated then EloTextBox.Repaint;
end;

procedure TEloStringsBox.LoadFromStream(Stream: TStream);
begin
  NoChanged:= True;
  inherited LoadFromStream(Stream);
  if EloTextBox.HandleAllocated then
  begin
   EloTextBox.ResetPosParams;
   EloTextBox.SetTextVScrollPos(EloTextBox.Handle, SB_VERT, 0, True);
  end;
  NoChanged:= False;
  Changed;
end;

procedure TEloStringsBox.Assign(Source: TPersistent);
begin
  NoChanged:= True;
  inherited Assign(Source);
  if EloTextBox.HandleAllocated then
  begin
   EloTextBox.ResetPosParams;
   EloTextBox.SetTextVScrollPos(EloTextBox.Handle, SB_VERT, 0, True);
  end;
  NoChanged:= False;
  Changed;
end;

{ TCustomEloTextBox }

constructor TCustomEloTextBox.Create(AOwner: TComponent);
const
  EditStyle = [csClickEvents, csSetCaption, csDoubleClicks, csFixedHeight];
begin
  inherited Create(AOwner);
  if NewStyleControls then
     ControlStyle := EditStyle
  else ControlStyle := EditStyle + [csFramed];
  AutoSize:= False;
  FVScrollBar:= False;
  ParagraphArray:= nil;
  FStrings:= TEloStringsBox.Create;
  TEloStringsBox(FStrings).EloTextBox:= Self;
  WorkList:= TStringList.Create;
  FTabString:= '   ';
  FAlignmentText:= atWidth;
  ParentColor:= False;
  ParentBackground:= False;
  Color:= clWindow;
  BevelKind:= bkSoft;
  BevelInner:= bvNone;
  BevelOuter:= bvLowered;
  FOEMConvert:= tcNone;
  FSpacing:= 4;
  FHighlightColor:= clHighlight;
  FHighlightTextColor:= clHighlightText;
  FHotTrack:= True;
  RangeBlockingMouseMove:= 500;
end;

procedure TCustomEloTextBox.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  if FVScrollBar then with Params do Style:= Style or WS_VSCROLL;
end;

destructor TCustomEloTextBox.Destroy;
begin
  ParagraphArray:= nil;
  FreeAndNil(WorkList);
  FreeAndNil(FStrings);
  inherited Destroy;
end;

procedure TCustomEloTextBox.DefaultHandler(var Message);
var
   scrollDelta, MouseYPos: Integer;
begin
  case TMessage(Message).Msg of
    WM_LBUTTONDOWN, WM_RBUTTONDOWN, WM_MBUTTONDOWN:
      begin
       if not Focused and Visible and Enabled then SetFocus;
       inherited DefaultHandler(Message);
      end;
    WM_MOUSEMOVE:
      begin
       if (Canvas.LockCount< 1)
       and (GetTickCount- BeginBlockingMouseMove> RangeBlockingMouseMove)
       and FHotTrack then
       begin
        MouseYPos:= TWMMouse(Message).YPos;
        if (LineHeight> 0)
        and (WorkList.Count> 0)
        and (MouseYPos>= 0) then
        begin
         OldLine:= CurrentLine;
         CurrentLine:= MouseYPos div LineHeight;
         if CurrentLine<> OldLine then
         begin
          PaintAllLines:= False;
          Repaint;
          PaintAllLines:= True;
         end;
        end;
       end;
       inherited DefaultHandler(Message);
      end;
    WM_MOUSEWHEEL:
      begin
       TMessage(Message).Result:= 1;
       if TWMMouseWheel(Message).WheelDelta< 0 then scrollDelta:= 1
       else scrollDelta:= 0;
       Perform(WM_VSCROLL, scrollDelta, 0);
       inherited DefaultHandler(Message);
      end;
    WM_SETFOCUS:
      begin
       if (Win32Platform = VER_PLATFORM_WIN32_WINDOWS)
       and not IsWindow(TWMSetFocus(Message).FocusedWnd) then
           TWMSetFocus(Message).FocusedWnd:= 0;
       inherited DefaultHandler(Message);
      end;
    WM_KILLFOCUS:
      begin
       if FHotTrack and ((CurrentLine> -1) or (OldLine> -1)) then
       begin
        CurrentLine:= -1;
        OldLine:= -1;
        PaintAllLines:= False;
        Repaint;
        PaintAllLines:= True;
       end;
       inherited DefaultHandler(Message);
      end;
    else inherited DefaultHandler(Message);
  end;
end;

procedure TCustomEloTextBox.SetAlignmentText(Value: TAlignmentText);
begin
  if Value<> FAlignmentText then
  begin
   FAlignmentText:= Value;
   ResetPosParams;
   Repaint;
  end;
end;

procedure TCustomEloTextBox.SetStrings(Value: TStrings);
begin
  FStrings.Assign(Value);
  ResetPosParams;
  SetTextVScrollPos(Handle, SB_VERT, 0, True);
  Repaint;
end;

procedure TCustomEloTextBox.SetTabString(Value: String);
begin
  if Value<> FTabString then
  begin
   if Value= '' then Value:= ' ';
   FTabString:= Value;
   Repaint;
  end;
end;

procedure TCustomEloTextBox.SetFileName(Value: TFileName);
begin
  if value<> FFileName then
  begin
   FFileName:= Value;
   ResetPosParams;
   SetTextVScrollPos(Handle, SB_VERT, 0, True);
   Repaint;
  end;
end;

procedure TCustomEloTextBox.SetVScrollBar(Value: Boolean);
begin
  if FVScrollBar<> Value then
  begin
   FVScrollBar:= Value;
   RecreateWnd;
  end;
end;

procedure TCustomEloTextBox.SetOEMConvert(Value: TOEMConvert);
begin
  if Value<> FOEMConvert then
  begin
   FOEMConvert:= Value;
   Repaint;
  end;
end;

procedure TCustomEloTextBox.SetSpacing(Value: Byte);
begin
  if Value<> FSpacing then
  begin
   FSpacing:= Value;
   Repaint;
  end;
end;

procedure TCustomEloTextBox.SetHighlightColor(Value: TColor);
begin
  if Value<> FHighlightColor then FHighlightColor:= Value;
end;

procedure TCustomEloTextBox.SetHighlightTextColor(Value: TColor);
begin
  if Value<> FHighlightTextColor then FHighlightTextColor:= Value;
end;

procedure TCustomEloTextBox.SetHotTrack(Value: Boolean);
begin
  if Value<> FHotTrack then FHotTrack:= Value;
end;

function TCustomEloTextBox.GetCurHeight: Integer;
var
   delta: Integer;
begin
  delta:= 0;
  if BevelKind<> bkNone then
  begin
   if BevelInner<> bvNone then
   begin
    if (beTop in BevelEdges) then delta:= delta+ BevelWidth;
    if (beBottom in BevelEdges) then delta:= delta+ BevelWidth;
   end;
   if BevelOuter<> bvNone then
   begin
    if (beTop in BevelEdges) then delta:= delta+ BevelWidth;
    if (beBottom in BevelEdges) then delta:= delta+ BevelWidth;
   end;
  end;
  delta:= delta+ BorderWidth* 2;
  Result:= Height- delta;
end;

function TCustomEloTextBox.GetCurWidth: Integer;
var
   delta: Integer;
begin
  delta:= 0;
  if BevelKind<> bkNone then
  begin
   if BevelInner<> bvNone then
   begin
    if (beLeft in BevelEdges) then delta:= delta+ BevelWidth;
    if (beRight in BevelEdges) then delta:= delta+ BevelWidth;
   end;
   if BevelOuter<> bvNone then
   begin
    if (beLeft in BevelEdges) then delta:= delta+ BevelWidth;
    if (beRight in BevelEdges) then delta:= delta+ BevelWidth;
   end;
  end;
  delta:= delta+ BorderWidth* 2;
  if FVScrollBar then delta:= delta+ GetSystemMetrics(SM_CXHSCROLL)+ FSpacing;
  Result:= Width- delta;
end;

{
  ,     Wd,    .
,  Wd     ,   .
      a  b.   a - 
   ,  b -      .



      
   ababaabbbaa   aaabaa

-         , 
 .  ,    .   -
  .
}
function TCustomEloTextBox.RusHyps(Wd: String): String;
const
     Vowel=['', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', ''];
     Symbs=['', '', '', ''];
     Cons=[''..'']- Vowel- Symbs;
var
   I: Integer;
   VW: Boolean;
begin
  if FAlignmentText= atWidthAndHyphen then
  begin
   VW:= False;
   Result:= '';
   I:= 1;
   repeat
    if (Wd[I] in Vowel) and (length(Result)> 0) then
    begin
     Result[length(Result)]:= 'b';
     VW:= True;
    end;
    while (I< length(Wd)) and (Wd[I+ 1] in Cons) do
    begin
     Result:= Result+ 'a';
     Inc(I);
    end;
    if I= length(Wd) then Break
    else if Wd[I+ 1] in Vowel then
         begin
          if VW and (length(Result)> 1) then Result[length(Result)]:= 'b'
          else VW:=True;
          Result:= Result+ 'aa';
          Inc(I, 2);
         end
         else if Wd[I+ 1] in Symbs then
                 if I+ 1=length(Wd) then
                 begin
                  Result:= Result+ 'aa';
                  Inc(I)
                 end
                 else if Wd[I+ 2] in Vowel then
                      begin
                       if VW and (length(Result)> 1) then Result[length(Result)]:= 'b'
                       else VW:= True;
                       Result:= Result+ 'aaa';
                       Inc(I, 3)
                      end
                      else
                      begin
                       Result:= Result+ 'aa';
                       Inc(I, 2)
                      end
              else
              begin
               Result:= Result+ 'a';
               Inc(I);
              end;
   until I>= length(Wd);
   if length(Wd)<> length(Result) then Result:= Result+ 'a';
  end // end if ChkBoxHyp.Checked then
  else
  begin
   Result:= '';
   for I:= 1 to length(Wd) do Result:= Result+ 'a';
  end
end;

{
     ,    .
 -  ,  "   ",      
  ,      ,  
,   -.      
          RusHyps,  
   .     
    ,  .    
,         "". ,
Hyphen('-')='aaacabaaa'.
}
function TCustomEloTextBox.Hyphen(Wd: String): String;
var
   RWd: String;
   I: Integer;
begin
  RWd:= '';
  Result:= '';
  for I:= 1 to length(Wd) do
      case Wd[I] of
       '-': begin
             if RWd<> '' then
             begin
              Result:= Result+ RusHyps(RWd);
              RWd:= '';
             end;
             Result:= Result+ 'c'
            end;
       ''..'': RWd:= RWd+ Wd[I];
       else begin
             if RWd<> '' then
             begin
              Result:= Result+ RusHyps(RWd);
              RWd:= '';
             end;
             Result:= Result+ 'a';
            end;
      end; // end case Wd[I] of
  if RWd<> '' then Result:= Result+ RusHyps(RWd);
end;

{
        ,    ,
 ,      ,    
    .  EmptyAllowed  False 
  ,   -   .   , -,  
   ,  -,      
   ,        ,
 - "   ".
}
function TCustomEloTextBox.GetWordPart(Wd: String; ML: Integer;
  EmptyAllowed: Boolean; var J: Integer): String;
var
   I: Integer;
   Hp, Pref: String;
begin
  Hp:= Hyphen(Wd);
  if EmptyAllowed then Pref:= ' '
  else Pref:= '';

  for I:= length(Hp) downto 1 do
      case Hp[I] of
       'b': if Canvas.TextWidth(Pref+ Copy(Wd, 1, I)+ '-')<= ML then
            begin
             Result:= Pref+ Copy(Wd, 1, I)+ '-';
             J:= I;
             exit;
            end;
       'c': if Canvas.TextWidth(Pref+ Copy(Wd, 1, I))<= ML then
            begin
             Result:= Pref+ Copy(Wd, 1, I);
             J:= I;
             exit;
            end;
      end;
  Result:= '';
  J:= 0;
  if not EmptyAllowed then
  begin
   for I:= 1 to length(Hp) do
       case Hp[I] of
        'b': begin
              Result:= Copy(Wd, 1, I)+ '-';
              J:= I;
              exit;
             end;
        'c': begin
              Result:= Copy(Wd, 1, I);
              J:= I;
              exit;
             end;
       end;
   Result:= Wd;
   J:= length(Wd);
  end;
end;

function TCustomEloTextBox.CheckString(strSource: String): String;
var
   i: Integer;
begin
  Result:= copy(strSource, 1, length(strSource));

  while (length(Result)> 0) and (Result[1]= ' ') do delete(Result, 1, 1);
  while (length(Result)> 0) and (Result[length(Result)]= ' ') do delete(Result, length(Result), 1);
  while pos('  ', Result)> 0 do delete(Result, pos('  ', Result), 1);

  if length(Result)> 0 then
  begin
   i:= 1;
   while i< length(Result) do
   begin
    if Result[i]= #9 then
    begin
     Insert(FTabString, Result, i);
     Delete(Result, i+ length(FTabString), 1);
     i:= i+ length(FTabString)- 1;
    end
    else if Result[i]< ' ' then
         begin
          Delete(Result, i, 1);
          if i> 1 then Dec(i);
         end;
    inc(i);
   end;
  end;
end;

procedure TCustomEloTextBox.MakeList(List: TStringList);
var
   n, I, J, K: Integer;
   strSource, S: String;
   curWidth: Integer;

begin
  List.Clear;
  if (ParagraphArray= nil) or (length(ParagraphArray)< 1) then exit;
  curWidth:= GetCurWidth;
  for n:= Low(ParagraphArray) to High(ParagraphArray) do
  begin
   strSource:= CheckString(ParagraphArray[n]);
   I:= 1;
   repeat
    J:= I+ 1;
    //        
    //   
    //        
    while (J<= length(strSource)) and (strSource[J]<> ' ') do Inc(J);
    //     .
    //       EloTextBox' -  .
    if Canvas.TextWidth(Copy(strSource, I, J- I))<= curWidth then
       S:= Copy(strSource, I, J- I)
    //   -      
    else
    begin
     S:= GetWordPart(Copy(strSource, I, J- I), curWidth, False, K);
     Inc(J, K);
    end;
    I:=J+ 1;
    //        
    while I<= length(strSource) do
    begin
     J:= I+ 1;
     while (J<= length(strSource)) and (strSource[J]<> ' ') do Inc(J);
     if Canvas.TextWidth(S+ ' '+ Copy(strSource, I, J- I))< curWidth then
        S:= S+ ' '+ Copy(strSource, I, J- I)
     else
     begin
      S:= S+ GetWordPart(Copy(strSource, I, J- I), curWidth- Canvas.TextWidth(S), True, K);
      I:= I+ K;
      Break;
     end;
     I:= J+ 1;
    end;
    //  
    List.Add(S);
   //    ,     .
   until I> length(strSource)
  end; // end else if FAlignmentText= atNone
end;

procedure TCustomEloTextBox.LoadFromFile(FileName: TFileName);
var
   S: String;
   F: TextFile;
   i: Integer;
begin
  //    
  if (FileName= '') or not FileExists(FileName) then
  begin
   MessageDlg(':    ', mtError, [mbCancel], 0);
   FFileName:= '';
   exit;
  end;
  ParagraphArray:= nil;
  AssignFile(F, FileName);
  Reset(F);
  i:= 0;
  Setlength(ParagraphArray, 1);
  while not EOF(F) do
  begin
   ReadLn(F, S);
   if Trim(S)<>'' then ParagraphArray[i]:= ParagraphArray[i]+ S+ ' '
   else
   begin
    Inc(i);
    Setlength(ParagraphArray, i+ 1);
   end;
  end;
  CloseFile(F);
end;

procedure TCustomEloTextBox.LoadFromList(List: TStrings);
var
   i, n: Integer;
begin
  //   List
  ParagraphArray:= nil;
  if List.Count< 1 then exit;
  n:= 0;
  Setlength(ParagraphArray, 1);
  for i:= 0 to List.Count- 1 do
  begin
   if Trim(List.Strings[i])<>'' then
      ParagraphArray[n]:= ParagraphArray[n]+ List.Strings[i]+ ' '
   else
   begin
    inc(n);
    Setlength(ParagraphArray, n+ 1);
   end;
  end;
end;

procedure TCustomEloTextBox.SetTextScrollRange(HWND: THandle; nBar, nMinPos, nMaxPos: Integer;
  bRedraw: Boolean);
begin
  if FVScrollBar then SetScrollRange(HWND, nBar, nMinPos, nMaxPos, bRedraw);
end;

procedure TCustomEloTextBox.SetTextVScrollPos(HWND: THandle; nBar, nPos: Integer;
  bRedraw: Boolean);
begin
  if FVScrollBar then SetScrollPos(HWND, nBar, nPos, bRedraw);
end;

procedure TCustomEloTextBox.ResetPosParams;
begin
  BeginScrollPos:= 0;
  VisibleLineCount:= 0;
  CurrentLine:= -1;
  OldLine:= -1;
end;

procedure TCustomEloTextBox.BlockingMouseMove;
begin
  BeginBlockingMouseMove:= GetTickCount;
end;

procedure TCustomEloTextBox.WMErasebkgnd(var Message: TWMErasebkgnd);
begin
  Message.Result:= 1;
end;

procedure TCustomEloTextBox.DrawCurrentLine(aCanvas: TCanvas; X, Y: Integer;
  aRect: TRect; var Text: String);
var
   defDraw: Boolean;

   procedure CheckErrorFormatString;
   begin
     if (Trim(Text)<> '')
     and (X+ aCanvas.TextWidth(Text)> GetCurWidth) then
         if Assigned(FOnErrorTextFormat) then
            FOnErrorTextFormat(Self, aCanvas, Text);
   end;

   procedure DefaultDrawText;
   begin
     if IndexLineClear<> Y then
     begin
      aCanvas.FillRect(aRect);
      IndexLineClear:= Y;
     end;
     aCanvas.TextOut(X, Y, Text);
   end;

   procedure DefaultDrawLine;
   begin
     if not Enabled then
     begin
      aCanvas.Font.Color:= clGrayText;
      aCanvas.Brush.Color:= Color;
     end
     else
     begin
      aCanvas.Font.Color:= Font.Color;
      if not EventMouseLeave
      and (CurrentLine> -1)
      and (Y= LineHeight* CurrentLine) then
      begin
       aCanvas.Font.Color:= FHighlightTextColor;
       aCanvas.Brush.Color:= FHighlightColor;
      end
      else
      begin
       aCanvas.Font.Color:= Font.Color;
       aCanvas.Brush.Color:= Color;
      end;
     end;
     if PaintAllLines
     or (Y= LineHeight* CurrentLine)
     or (Y= LineHeight* OldLine) then DefaultDrawText;
     aCanvas.Font.Color:= Font.Color;
     aCanvas.Brush.Color:= Color;
   end;

begin
  case FOEMConvert of
    tcAnsiToOem: Text:= ConvertAnsiToOem(Text);
    tcOemToAnsi: Text:= ConvertOemToAnsi(Text);
  end;
  CheckErrorFormatString;
  defDraw:= True;
  if Assigned(FOnDrawCurrentLine) then
     FOnDrawCurrentLine(aCanvas, X, Y, aRect, Text, defDraw);
  if defDraw then DefaultDrawLine;
end;

procedure TCustomEloTextBox.Paint;
var
   I, J, W, Spaces, SpSize, SpW, X,
   beginIndex, endIndex, curWidth, FirstIndex: Integer;
   BeginParagraph, EndParagraph: Boolean;
   curStr, curWord: String;
   PaintRect: TRect;
begin
  inherited;
  IndexLineClear:= -1;
  with Canvas do
  begin
   Font:= Self.Font;
   Brush.Color:= Color;
   Brush.Style:= bsSolid;
   Pen.Style:= psSolid;

   if PaintAllLines then
   begin
    //  
    if FFileName<> '' then LoadFromFile(FFileName)
    else LoadFromList(FStrings);
    MakeList(TStringList(WorkList));
   end;
   curWidth:= GetCurWidth;

   //  
   LineHeight:= TextHeight('');
   VisibleLineCount:= (GetCurHeight div LineHeight);
   if FVScrollBar then SetTextScrollRange(Self.Handle, 1, 0, WorkList.Count- VisibleLineCount, True);

   if WorkList.Count< 1 then
   begin
    FillRect(ClientRect);
    SetVScrollBar(False);
    LineHeight:= 0;
    ResetPosParams;
    FBOFText:= True;
    FEOFText:= True;
    if Assigned(FOnPaint) then FOnPaint(Self);
    exit;
   end;

   if FVScrollBar<> (VisibleLineCount<= WorkList.Count) then
   begin
    SetVScrollBar(VisibleLineCount<= WorkList.Count);
    if Assigned(FOnPaint) then FOnPaint(Self);
    exit;
   end;

   if BeginScrollPos> WorkList.Count- VisibleLineCount then
      BeginScrollPos:= WorkList.Count- VisibleLineCount;
   if BeginScrollPos< 0 then BeginScrollPos:= 0;
   beginIndex:= BeginScrollPos;
   if VisibleLineCount< WorkList.Count then
      endIndex:= beginIndex+ VisibleLineCount
   else endIndex:= beginIndex+ WorkList.Count;
   if endIndex> WorkList.Count then endIndex:= WorkList.Count;

   FBOFText:= beginIndex<= 0;
   FEOFText:= endIndex>= WorkList.Count;

   endIndex:= endIndex- 1;

   for I:= beginIndex to endIndex do
   begin
    curStr:= WorkList[I];
    // 
    if pos(' ', curStr)> 0 then
       sleep(0);
    // 
    FirstIndex:= 1;
    X:= 0;
    PaintRect.Left:= ClientRect.Left;
    PaintRect.Top:= ClientRect.Top+ (I- beginIndex)* LineHeight;
    PaintRect.Right:= PaintRect.Left+ curWidth+ FSpacing;
    PaintRect.Bottom:= PaintRect.Top+ LineHeight;

    BeginParagraph:= copy(curStr, 1, length(FTabString))= FTabString;
    EndParagraph:= (WorkList.Count<= 1)
                   or (I>= WorkList.Count- 1)
                   or ((WorkList.Count> 1)
                      and ((copy(WorkList[I+ 1], 1, length(FTabString))= FTabString)
                          or (Trim(WorkList[I+ 1])= '')));

    if BeginParagraph then
    begin
     //   
     curStr:= copy(curStr, length(FTabString)+ 1, length(curStr));
     X:= TextWidth(FTabString);
     //   
     DrawCurrentLine(Canvas, 0, (I- beginIndex)* LineHeight, PaintRect, FTabString);
    end;
    if EndParagraph then
    begin
     //     ,
     //       .
     DrawCurrentLine(Canvas, X, (I- beginIndex)* LineHeight, PaintRect, curStr);
     continue;
    end;

    Spaces:= 0;
    //     
    for J:= FirstIndex to length(curStr)- 1 do
        if curStr[J]= ' ' then Inc(Spaces);
    //   ,    
    //   EloTextBox',    
    if (FAlignmentText= atNone)
    or (Spaces= 0)
    or (TextWidth(curStr)= curWidth- X) then
       DrawCurrentLine(Canvas, X, (I- beginIndex)* LineHeight, PaintRect, curStr)
    else
    begin
     //       ,
     //        ,  
     SpSize:= curWidth- X;
     W:= 1;
     for J:= FirstIndex to length(curStr)- 1 do
         if curStr[J]= ' ' then
         begin
          Dec(SpSize, TextWidth(Copy(curStr, W, J- W)));
          W:= J+ 1;
         end;
     Dec(SpSize, TextWidth(Copy(curStr, W, MaxInt)));
     {   TextWidth(' Delphi')
          
       TextWidth('')+ TextWidth(' ')+ TextWidth('Delphi)}

     //     ,
     //    ,   .
     W:= 1;
     while Spaces> 0 do
     begin
      J:= W;
      while curStr[J]<> ' ' do Inc(J);
      curWord:= Copy(curStr, W, J- W);
      DrawCurrentLine(Canvas, X, (I- beginIndex)* LineHeight, PaintRect, curWord);
      Inc(X, TextWidth(Copy(curStr, W, J- W)));
      SpW:= Round(SpSize/ Spaces);
      Inc(X, SpW);
      Dec(SpSize, SpW);
      Dec(Spaces);
      W:= J+ 1;
     end;

     curWord:= Copy(curStr, W,length(curStr));
     DrawCurrentLine(Canvas, X, (I- beginIndex)* LineHeight, PaintRect, curWord);
    end;
   end;
   //    
   if (endIndex- beginIndex+ 1)* LineHeight< Height then
      FillRect(Rect(ClientRect.Left, (endIndex- beginIndex+ 1)* LineHeight, ClientRect.Right, ClientRect.Bottom));
  end;
  IndexLineClear:= -1;
  if Assigned(FOnPaint) then FOnPaint(Self);
end;

procedure TCustomEloTextBox.Clear;
begin
  FFileName:= '';
  FStrings.Clear;
  ResetPosParams;
  SetTextVScrollPos(Handle, SB_VERT, 0, True);
end;

procedure TCustomEloTextBox.WMWindowPosChanged(var Message: TWMWindowPosChanged);
begin
  PaintAllLines:= True;
  BlockingMouseMove;
  inherited;
  BlockingMouseMove;
end;

procedure TCustomEloTextBox.WMSize(var Message: TWMSize);
begin
  PaintAllLines:= True;
  BlockingMouseMove;
  inherited;
  BlockingMouseMove;
  Canvas.FillRect(ClientRect);
  Invalidate;
end;

procedure TCustomEloTextBox.WMNCPaint(var Message: TMessage);
begin
  BlockingMouseMove;
  inherited;
  PaintAllLines:= True;
  BlockingMouseMove;
  Canvas.FillRect(ClientRect);
  Invalidate;
end;

procedure TCustomEloTextBox.WMSetFont(var Message: TWMSetFont);
begin
  PaintAllLines:= True;
  BlockingMouseMove;
  Canvas.FillRect(ClientRect);
  inherited;
end;

procedure TCustomEloTextBox.WMVScroll(var Message: TWMVScroll);
var
   NewPos, RangeMin, RangeMax: Integer;
   ScrollPos: Integer;
begin
  //BlockingMouseMove;
  if not Focused and Visible and Enabled then SetFocus;
  GetScrollRange(Handle, SB_VERT, RangeMin, RangeMax);
  with Message do
  begin
   NewPos:= BeginScrollPos;
   case ScrollCode of
    SB_LINEUP:	Dec(NewPos, 1);
    SB_LINEDOWN:	Inc(NewPos, 1);
    SB_TOP: NewPos:= 0;
    SB_BOTTOM: NewPos:= RangeMax- VisibleLineCount;
    SB_PAGEUP: Dec(NewPos, VisibleLineCount- 1);
    SB_PAGEDOWN: Inc(NewPos, VisibleLineCount- 1);
    SB_THUMBPOSITION:;	
    SB_THUMBTRACK: NewPos:= Pos;
    SB_ENDSCROLL:;
   end;
   if NewPos> RangeMax then NewPos:= RangeMax;
   if NewPos< 0 then NewPos:= 0;
   if BeginScrollPos<> NewPos then
   begin
    BeginScrollPos:= NewPos;
    SetTextVScrollPos(Handle, SB_VERT, BeginScrollPos, True);
    Repaint;
   end;
  end;
  if Assigned(FOnVScroll) then
  begin
   ScrollPos:= Message.Pos;
   FOnVScroll(Self, TScrollCode(Message.ScrollCode), ScrollPos);
   Message.Pos:= ScrollPos;
  end;
end;

procedure TCustomEloTextBox.WMEnable(var Message: TWMEnable);
begin
  inherited;
  Repaint;
end;

procedure TCustomEloTextBox.CMMouseLeave(var Message: TMessage);
begin
  if FHotTrack and ((CurrentLine> -1) or (OldLine> -1)) then
  begin
   EventMouseLeave:= True;
   PaintAllLines:= False;
   Repaint;
   CurrentLine:= -1;
   OldLine:= -1;
   PaintAllLines:= True;
   EventMouseLeave:= False;
  end;
  inherited;
end;

procedure TCustomEloTextBox.CNKeyDown(var Message: TWMKeyDown);
var
   NewPos, RangeMin, RangeMax: Integer;
   ShiftState: TShiftState;
begin
  //BlockingMouseMove;
  if Message.CharCode= VK_TAB then inherited;
  Message.Result:= 1;
  if Visible and Enabled then
  begin
   GetScrollRange(Handle, SB_VERT, RangeMin, RangeMax);
   NewPos:= BeginScrollPos;
   case Message.CharCode of
     VK_UP: Dec(NewPos, 1);
     VK_DOWN: Inc(NewPos, 1);
     VK_PRIOR: Dec(NewPos, VisibleLineCount- 1);
     VK_NEXT: Inc(NewPos, VisibleLineCount- 1);
     VK_HOME: NewPos:= 0;
     VK_END: NewPos:= RangeMax;
   end;
   if NewPos> RangeMax then NewPos:= RangeMax;
   if NewPos< 0 then NewPos:= 0;
   if BeginScrollPos<> NewPos then
   begin
    BeginScrollPos:= NewPos;
    SetTextVScrollPos(Handle, SB_VERT, BeginScrollPos, True);
    Repaint;
   end;
  end;
  if Assigned(OnKeyDown) then
  begin
   ShiftState := KeyDataToShiftState(Message.KeyData);
   OnKeyDown(Self, Message.CharCode, ShiftState);
  end;
end;

procedure TCustomEloTextBox.TextHome;
begin
  PostMessage(Handle, WM_KEYDOWN, VK_HOME, 0);
end;

procedure TCustomEloTextBox.TextPageUp;
begin
  PostMessage(Handle, WM_KEYDOWN, VK_PRIOR, 0);
end;

procedure TCustomEloTextBox.TextPrior;
begin
  PostMessage(Handle, WM_KEYDOWN, VK_UP, 0);
end;

procedure TCustomEloTextBox.TextNext;
begin
  PostMessage(Handle, WM_KEYDOWN, VK_DOWN, 0);
end;

procedure TCustomEloTextBox.TextPageDown;
begin
  PostMessage(Handle, WM_KEYDOWN, VK_NEXT, 0);
end;

procedure TCustomEloTextBox.TextEnd;
begin
  PostMessage(Handle, WM_KEYDOWN, VK_END, 0);
end;

end.

